/*****************************************************************************
 * rist.c: RIST (Reliable Internet Stream Transport) input module
 *****************************************************************************
 * Copyright (C) 2018, DVEO, the Broadcast Division of Computer Modules, Inc.
 * Copyright (C) 2018-2020, SipRadius LLC
 *
 * Authors: Sergio Ammirata <sergio@ammirata.net>
 *          Daniele Lacamera <root@danielinux.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_interrupt.h>
#include <vlc_plugin.h>
#include <vlc_access.h>
#include <vlc_threads.h>
#include <vlc_block.h>
#include <vlc_url.h>

#include "rist.h"

typedef struct
{
    struct       rist_server *server_ctx;
    bool         b_flag_discontinuity;
    int          gre_filter_src_port;
    int          gre_filter_dst_port;
    block_fifo_t *p_fifo;
} stream_sys_t;

static void cb_recv(void *arg, struct rist_peer *peer, uint64_t flow_id, const void *buf, size_t len, uint16_t src_port, uint16_t dst_port, uint64_t timestamp_ntp, uint32_t i_flags)
{
    stream_t *p_access = (stream_t *)arg;
    stream_sys_t *p_sys = p_access->p_sys;
    (void) flow_id;
    (void) peer;

    if (p_sys->gre_filter_src_port != 0 && p_sys->gre_filter_src_port != src_port)
        msg_Err( p_access, "Source port mismatch %d != %d\n", p_sys->gre_filter_src_port, src_port);

    if (p_sys->gre_filter_dst_port != 0 && p_sys->gre_filter_dst_port != dst_port)
        msg_Err( p_access, "Destination port mismatch %d != %d\n", p_sys->gre_filter_dst_port, dst_port);

    block_t *pkt = block_Alloc(len);
    if (pkt)
    {
        memcpy(pkt->p_buffer, buf, len);
        pkt->i_buffer = len;
        // Convert ntp64 to vlc_ticks
        uint64_t nanoseconds = (uint64_t)(timestamp_ntp >> 32);
        nanoseconds *= 1000000000;
        nanoseconds += timestamp_ntp & 0x000000ffffffff;
        pkt->i_dts = pkt->i_pts = VLC_TICK_FROM_US( nanoseconds / 1000 );
        pkt->i_flags = i_flags;
        block_FifoPut( p_sys->p_fifo, pkt );
    }
    else {
        msg_Err( p_access, "cb_recv OOM!" );
    }
}

static int Control(stream_t *p_access, int i_query, va_list args)
{
    switch( i_query )
    {
        case STREAM_CAN_SEEK:
        case STREAM_CAN_FASTSEEK:
        case STREAM_CAN_PAUSE:
        case STREAM_CAN_CONTROL_PACE:
            *va_arg( args, bool * ) = false;
            break;

        case STREAM_GET_PTS_DELAY:
            *va_arg( args, vlc_tick_t * ) = VLC_TICK_FROM_MS(
                   var_InheritInteger(p_access, "network-caching") );
            break;

        default:
            return VLC_EGENERIC;
    }

    return VLC_SUCCESS;
}

static block_t *BlockRIST(stream_t *p_access, bool *restrict eof)
{
    stream_sys_t *p_sys = p_access->p_sys;
    *eof = false;

    if (vlc_killed())
    {
        *eof = true;
        return NULL;
    }

    block_t *pktout = block_FifoGet(p_sys->p_fifo);
    return pktout;

}

static void Clean(struct stream_t *p_access)
{
    stream_sys_t *p_sys = p_access->p_sys;

    if( likely(p_sys->p_fifo != NULL) )
        block_FifoRelease( p_sys->p_fifo );

    rist_server_shutdown(p_sys->server_ctx);
}

static void Close(vlc_object_t *p_this)
{
    stream_t *p_access = (stream_t*)p_this;
    Clean(p_access);
}

static int Open(vlc_object_t *p_this)
{
    stream_t     *p_access = (stream_t*)p_this;
    stream_sys_t *p_sys = NULL;
    char *addr[RIST_PEER_COUNT];

    p_sys = vlc_obj_calloc( p_this, 1, sizeof( *p_sys ) );
    if( unlikely( p_sys == NULL ) )
        return VLC_ENOMEM;

    p_access->p_sys = p_sys;

    p_sys->gre_filter_src_port = var_InheritInteger(p_access, "gre-src-port");
    p_sys->gre_filter_dst_port = var_InheritInteger(p_access, "gre-dst-port");
    if (p_sys->gre_filter_dst_port % 2 != 0) {
        msg_Err( p_access, "GRE destination port must be an even number." );
        return VLC_ENOMEM;
    }

    int i_recovery_length = var_InheritInteger( p_access, RIST_PARAM_BUFFER_SIZE );
    int i_rist_reorder_buffer = var_InheritInteger( p_access, RIST_PARAM_REORDER_BUFFER );
    int i_rist_retry_interval = var_InheritInteger( p_access, RIST_PARAM_RETRY_INTERVAL );
    int i_rist_max_retries = var_InheritInteger( p_access, RIST_PARAM_MAX_RETRIES );
    int i_rist_max_bitrate = var_InheritInteger(p_access, RIST_PARAM_MAX_BITRATE);

    msg_Info( p_access, "Setting queue latency to %d ms", i_recovery_length );

	struct rist_server *server_ctx;

    int rist_profile = var_InheritInteger(p_access, RIST_PARAM_PROFILE);
    if (rist_server_create(&server_ctx, rist_profile) != 0) {
        msg_Err( p_access, "Could not create rist server context" );
        goto failed;
    }

    const struct rist_peer_config default_peer_config = {
        .address = NULL,
        .gre_dst_port = RIST_DEFAULT_DST_PORT + 1,
        .recovery_mode = RIST_RECOVERY_MODE_TIME,
        .recovery_maxbitrate = (uint32_t)i_rist_max_bitrate,
        .recovery_maxbitrate_return = 0,
        .recovery_length_min = (uint32_t)i_recovery_length,
        .recovery_length_max = (uint32_t)i_recovery_length,
        .recover_reorder_buffer = (uint32_t)i_rist_reorder_buffer,
        .recovery_rtt_min = (uint32_t)i_rist_retry_interval,
        .recovery_rtt_max = (uint32_t)i_rist_retry_interval,
        .weight = 5,
        .bufferbloat_mode = RIST_BUFFER_BLOAT_MODE_NORMAL,
        .bufferbloat_limit = 7,
        .bufferbloat_hard_limit = (uint32_t)i_rist_max_retries
    };

    int i_verbose_level = var_InheritInteger( p_access, RIST_PARAM_VERBOSE_LEVEL );
    if (rist_server_init(server_ctx, &default_peer_config, i_verbose_level, NULL, NULL, NULL) == -1) {
        msg_Err( p_access, "Could not init rist server" );
        goto failed;
    }

    char *psz_stream_name = NULL;
    psz_stream_name = var_InheritString( p_access, RIST_PARAM_CNAME );
    if ( psz_stream_name != NULL && psz_stream_name[0] != '\0')
    {
        if (rist_server_set_cname(server_ctx, psz_stream_name, strlen(psz_stream_name)) != 0) {
            msg_Err( p_access, "Could not set the cname" );
            free(psz_stream_name);
            goto failed;
        }
        free( psz_stream_name );
    }

    /* Enable encryption */
    char *psz_shared_secret = NULL;
    psz_shared_secret = var_InheritString( p_access, RIST_PARAM_SHARED_SECRET );
    if (psz_shared_secret != NULL && psz_shared_secret[0] != '\0') {
        int i_keysize = var_InheritInteger( p_access, RIST_PARAM_ENCRYPTION_TYPE );
        if (rist_server_encrypt_enable(server_ctx, psz_shared_secret, i_keysize) == -1) {
                msg_Err( p_access, "Could not enable encryption" );
                free(psz_shared_secret);
                goto failed;
        }
        free(psz_shared_secret);
    }

    addr[0] = strdup(p_access->psz_url);
    addr[1] = var_InheritString( p_access, RIST_PARAM_URL2 );
    addr[2] = var_InheritString( p_access, RIST_PARAM_URL3 );
    addr[3] = var_InheritString( p_access, RIST_PARAM_URL4 );
    bool b_peer_alive = false;
    for (size_t i = 0; i < RIST_PEER_COUNT; i++) {
        if (&addr[i] == NULL) {
            continue;
        }
        else if (addr[i][0] == '\0') {
            free(addr[i]);
            continue;
        }
        if (rist_server_add_peer(server_ctx, addr[i]) == -1) {
            msg_Err( p_access, "Could not init rist server%i at %s", (int)(i + 1), addr[i] );
            free(addr[i]);
            continue;
        }
        else {
            b_peer_alive = true;
        }
        free(addr[i]);
    }
    if (!b_peer_alive)
        goto failed;

    p_sys->p_fifo = block_FifoNew();
    if( unlikely(p_sys->p_fifo == NULL) )
        goto failed;

    /* Start the rist protocol thread */
    if (rist_server_start(server_ctx, cb_recv, p_access)) {
        msg_Err( p_access, "Could not start rist server" );
        goto failed;
    }

    p_access->pf_block = BlockRIST;
    p_access->pf_control = Control;

    return VLC_SUCCESS;

failed:
    Clean( p_access );
    return VLC_EGENERIC;
}

/* Module descriptor */
vlc_module_begin ()

    set_shortname( N_("RIST") )
    set_description( N_("RIST input") )
    set_category( CAT_INPUT )
    set_subcategory( SUBCAT_INPUT_ACCESS )

    add_string( RIST_PARAM_URL2, NULL, RIST_URL2_TEXT, NULL, true )
    add_string( RIST_PARAM_URL3, NULL, RIST_URL3_TEXT, NULL, true )
    add_string( RIST_PARAM_URL4, NULL, RIST_URL4_TEXT, NULL, true )
    add_integer( RIST_PARAM_MAX_BITRATE, RIST_DEFAULT_MAX_BITRATE,
            RIST_MAX_BITRATE_TEXT, RIST_MAX_BITRATE_LONGTEXT, true )
    add_integer( RIST_PARAM_RETRY_INTERVAL, RIST_DEFAULT_RETRY_INTERVAL, 
        RIST_RETRY_INTERVAL_TEXT, NULL, true )
    add_integer( RIST_PARAM_REORDER_BUFFER, RIST_DEFAULT_REORDER_BUFFER, 
        RIST_REORDER_BUFFER_TEXT, NULL, true )
    add_integer( RIST_PARAM_MAX_RETRIES, RIST_DEFAULT_MAX_RETRIES, 
        RIST_MAX_RETRIES_TEXT, NULL, true )
    add_integer( RIST_PARAM_VERBOSE_LEVEL, RIST_DEFAULT_VERBOSE_LEVEL,
            RIST_VERBOSE_LEVEL_TEXT, RIST_VERBOSE_LEVEL_LONGTEXT, true )
        change_integer_list( verbose_level_type, verbose_level_type_names )
    add_integer( RIST_PARAM_BUFFER_SIZE, RIST_DEFAULT_BUFFER_SIZE,
            BUFFER_TEXT, BUFFER_LONGTEXT, true )
    add_string( RIST_PARAM_CNAME, NULL, RIST_CNAME_TEXT, 
            RIST_CNAME_LONGTEXT, true )
    add_integer( RIST_PARAM_PROFILE, RIST_DEFAULT_PROFILE,
            RIST_PROFILE_TEXT, RIST_PROFILE_LONGTEXT, true )
    add_password( RIST_PARAM_SHARED_SECRET, "",
            RIST_SHARED_SECRET_TEXT, NULL )
    add_integer( RIST_PARAM_ENCRYPTION_TYPE, RIST_DEFAULT_ENCRYPTION_TPE,
            RIST_ENCRYPTION_TYPE_TEXT, NULL, true )
        change_integer_list( rist_encryption_type, rist_encryption_type_names )

    set_capability( "access", 0 )
    add_shortcut( "rist", "tr06" )

    set_callbacks( Open, Close )

vlc_module_end ()
