/*****************************************************************************
 * rist.h: RIST (Reliable Internet Stream Transport) helper
 *****************************************************************************
 * Copyright (C) 2018, DVEO, the Broadcast Division of Computer Modules, Inc.
 * Copyright (C) 2018-2020, SipRadius LLC
 *
 * Authors: Sergio Ammirata <sergio@ammirata.net>
 *          Daniele Lacamera <root@danielinux.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#ifndef RIST_COMMON_H
#define RIST_COMMON_H 1

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <librist.h>

/* RIST parameter names */
#define RIST_PARAM_SHARED_SECRET     "shared-secret"
#define RIST_PARAM_ENCRYPTION_TYPE   "encryption-type"
#define RIST_PARAM_BUFFER_SIZE       "buffer-size"
#define RIST_PARAM_PROFILE           "profile"
#define RIST_PARAM_CNAME             "cname"
#define RIST_PARAM_VERBOSE_LEVEL     "verbose-level"
#define RIST_PARAM_RETRY_INTERVAL    "retry-interval"
#define RIST_PARAM_REORDER_BUFFER    "reorder-buffer"
#define RIST_PARAM_MAX_RETRIES       "max-retries"
#define RIST_PARAM_MAX_BITRATE       "max-bitrate"
#define RIST_PARAM_URL2              "peer-url2"
#define RIST_PARAM_URL3              "peer-url3"
#define RIST_PARAM_URL4              "peer-url4"

/* RIST parameter descriptions */
#define RIST_URL2_TEXT N_("Second peer URL")
#define RIST_URL3_TEXT N_("Third peer URL")
#define RIST_URL4_TEXT N_("Fourth peer URL")
#define RIST_MAX_BITRATE_TEXT N_("Max bitrate in Kbps")
#define RIST_MAX_BITRATE_LONGTEXT N_( \
    "Use this value to guarantee that data+retries bitrate never exceeds your pipe size. " \
    "Default value is 100000 Kbps (100 Mbps)" )
#define RIST_RETRY_INTERVAL_TEXT N_("RIST nack retry interval (ms)")
#define RIST_REORDER_BUFFER_TEXT N_("RIST reorder buffer (ms)")
#define RIST_MAX_RETRIES_TEXT N_("RIST maximum retry count")
#define BUFFER_TEXT N_("RIST retry-buffer queue size (ms)")
#define BUFFER_LONGTEXT N_( \
    "This must match the buffer size (latency) configured on the other side. If you " \
    "are not sure, leave it blank and it will use 1000ms" )
#define RIST_SHARED_SECRET_TEXT N_("Shared Secret")
#define RIST_SHARED_SECRET_LONGTEXT N_( \
    "This shared secret is a passphare shared between sender and receiver. The AES key " \
    "is derived from it" )
#define RIST_CNAME_TEXT N_("Peer cname")
#define RIST_CNAME_LONGTEXT N_( \
    "This name will be sent using the rist RTCP channel and uniquely identifies us" )
/* Profile selection */
#define RIST_PROFILE_TEXT N_("Rist Profile")
#define RIST_PROFILE_LONGTEXT N_( "Select the rist profile to use" )
static const int rist_profile[] = { 0, 1, 2, };
static const char *const rist_profile_names[] = {
    N_("Simple Profile"), N_("Main Profile"), N_("Advanced Profile"),
};
/* Encryption type */
#define RIST_ENCRYPTION_TYPE_TEXT N_("Encryption type")
#define RIST_DEFAULT_ENCRYPTION_TPE 128
static const int rist_encryption_type[] = { 0, 128, 256, };
static const char *const rist_encryption_type_names[] = {
    N_("Disabled"), N_("AES 128 bits"), N_("AES 256 bits"),
};
/* Verbose level */
#define RIST_VERBOSE_LEVEL_TEXT N_("Verbose level")
#define RIST_VERBOSE_LEVEL_LONGTEXT N_("This controls how much log data the library will output over stderr")
static const int verbose_level_type[] = { RIST_LOG_QUIET, RIST_LOG_INFO, RIST_LOG_ERROR, RIST_LOG_WARN, RIST_LOG_DEBUG, RIST_LOG_SIMULATE, };
static const char *const verbose_level_type_names[] = {
    N_("Quiet"), N_("Info"), N_("Errors"),N_("Warnings"), N_("Debug"), N_("Simulate-Loss"),
};

/* Max stream-name length */
#define RIST_MAX_CNAME 128
/* Default buffer size in ms */
#define RIST_DEFAULT_BUFFER_SIZE 1000
/* Default verbose mode */
#define RIST_DEFAULT_VERBOSE_LEVEL RIST_LOG_WARN
/* The default nack retry interval */
#define RIST_DEFAULT_RETRY_INTERVAL 50
/* The default packet re-ordering buffer */
#define RIST_DEFAULT_REORDER_BUFFER 25
/* The default max bitrate in Kbps */
#define RIST_DEFAULT_MAX_BITRATE 100000
/* Default ports for main profile mode */
#define RIST_DEFAULT_SRC_PORT 1971
#define RIST_DEFAULT_DST_PORT 1968
/* Default profile */
#define RIST_DEFAULT_PROFILE RIST_PROFILE_MAIN
/* Default hard-stop for retries */
#define RIST_DEFAULT_MAX_RETRIES 20
/* Default retry spacing */
#define RIST_DEFAULT_RETRY_INTERVAL 50
/* The default packet re-ordering buffer */
#define RIST_DEFAULT_REORDER_BUFFER 25
/* Max number of peers */
#define RIST_PEER_COUNT 4

#endif
