/*****************************************************************************
 *  * rist.c: RIST (Reliable Internet Stream Transport) output module
 *****************************************************************************
 * Copyright (C) 2018, DVEO, the Broadcast Division of Computer Modules, Inc.
 * Copyright (C) 2018-2020, SipRadius LLC
 *
 * Authors: Sergio Ammirata <sergio@ammirata.net>
 *          Daniele Lacamera <root@danielinux.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_interrupt.h>
#include <vlc_plugin.h>
#include <vlc_sout.h>
#include <vlc_block.h>
#include <vlc_rand.h>
#include <sys/time.h>

#include "../access/rist.h"

#define SOUT_CFG_PREFIX "sout-rist-"

static const char *const ppsz_sout_options[] = {
    "multipeer-mode",
    "payload-size",
    "gre-src-port",
    "gre-dst-port",
    RIST_PARAM_MAX_BITRATE,
    RIST_PARAM_BUFFER_SIZE,
    RIST_PARAM_CNAME,
    RIST_PARAM_PROFILE,
    RIST_PARAM_ENCRYPTION_TYPE,
    RIST_PARAM_SHARED_SECRET,
    NULL
};

typedef struct
{
    struct rist_client *client_ctx;
    block_t            *p_pktbuffer;
    int                gre_src_port;
    int                gre_dst_port;
    bool               b_mtu_warning;
    size_t             i_packet_size;
} sout_access_out_sys_t;

static uint64_t i_dts_to_ntp64( vlc_tick_t i_dts )
{
    lldiv_t d = lldiv (i_dts, CLOCK_FREQ);
    struct timespec ts = { d.quot, NS_FROM_VLC_TICK( d.rem ) };
    // Convert nanoseconds to 32-bits fraction (232 picosecond units)
    uint64_t t = (uint64_t)(ts.tv_nsec) << 32;
    t /= 1000000000;
    // There is 70 years (incl. 17 leap ones) offset to the Unix Epoch.
    // No leap seconds during that period since they were not invented yet.
    t |= ((70LL * 365 + 17) * 24 * 60 * 60 + ts.tv_sec) << 32;
    return t; // nanoseconds (technically, 232.831 picosecond units)
}

static ssize_t Write( sout_access_out_t *p_access, block_t *p_buffer )
{
    sout_access_out_sys_t *p_sys = p_access->p_sys;
    int i_len = 0;

    while( p_buffer )
    {
        block_t *p_next;
        int i_block_split = 0;

        if( !p_sys->b_mtu_warning && p_buffer->i_buffer > p_sys->i_packet_size )
        {
            msg_Warn( p_access, "Buffer data size (%zu) > configured packet size (%zu), you " \
                "should probably increase the configured packet size", p_buffer->i_buffer, 
                p_sys->i_packet_size );
            p_sys->b_mtu_warning = true;
        }

        /* Temp buffer is already too large, flush */
        if( p_sys->p_pktbuffer->i_buffer + p_buffer->i_buffer > p_sys->i_packet_size )
        {
            rist_client_write_timed(p_sys->client_ctx, p_sys->p_pktbuffer->p_buffer, p_sys->p_pktbuffer->i_buffer,
                                    p_sys->gre_src_port, p_sys->gre_dst_port, i_dts_to_ntp64(p_sys->p_pktbuffer->i_dts));
            p_sys->p_pktbuffer->i_buffer = 0;
        }

        i_len += p_buffer->i_buffer;

        while( p_buffer->i_buffer )
        {

            size_t i_write = __MIN( p_buffer->i_buffer, p_sys->i_packet_size );

            i_block_split++;

            if( p_sys->p_pktbuffer->i_buffer == 0 )
            {
                p_sys->p_pktbuffer->i_dts = p_buffer->i_dts;
            }

            memcpy( p_sys->p_pktbuffer->p_buffer + p_sys->p_pktbuffer->i_buffer,
                    p_buffer->p_buffer, i_write );

            p_sys->p_pktbuffer->i_buffer += i_write;
            p_buffer->p_buffer += i_write;
            p_buffer->i_buffer -= i_write;

            /*  Flush if we reached the target size for the case of block size < target packet size.
             *  Also flush when we are in block_split > 1 for the case when the block_size is
             *  larger than the packet-size because we need to continue the inner loop */
            if( p_sys->p_pktbuffer->i_buffer == p_sys->i_packet_size || i_block_split > 1 )
            {
                rist_client_write_timed(p_sys->client_ctx, p_sys->p_pktbuffer->p_buffer, p_sys->p_pktbuffer->i_buffer,
                                        p_sys->gre_src_port, p_sys->gre_dst_port, i_dts_to_ntp64(p_sys->p_pktbuffer->i_dts));
                p_sys->p_pktbuffer->i_buffer = 0;
            }

        }

        p_next = p_buffer->p_next;
        block_Release( p_buffer );
        p_buffer = p_next;

    }

    if ( i_len <= 0 ) {
        block_ChainRelease( p_buffer );
    }
    return i_len;
}

static int Control( sout_access_out_t *p_access, int i_query, va_list args )
{
    VLC_UNUSED( p_access );

    int i_ret = VLC_SUCCESS;

    switch( i_query )
    {
        case ACCESS_OUT_CONTROLS_PACE:
            *va_arg( args, bool * ) = false;
            break;

        default:
            i_ret = VLC_EGENERIC;
            break;
    }

    return i_ret;
}

static void Close( vlc_object_t * p_this )
{
    sout_access_out_t     *p_access = (sout_access_out_t*)p_this;
    sout_access_out_sys_t *p_sys = p_access->p_sys;

    if (p_sys->p_pktbuffer)
        block_Release(p_sys->p_pktbuffer);

    rist_client_destroy(p_sys->client_ctx);
    p_sys->client_ctx = NULL;
}

static int Open( vlc_object_t *p_this )
{
    sout_access_out_t       *p_access = (sout_access_out_t*)p_this;
    sout_access_out_sys_t   *p_sys = NULL;
    char *addr[RIST_PEER_COUNT];

    if (var_Create ( p_access, "dst-port", VLC_VAR_INTEGER )
     || var_Create ( p_access, "src-port", VLC_VAR_INTEGER )
     || var_Create ( p_access, "dst-addr", VLC_VAR_STRING )
     || var_Create ( p_access, "src-addr", VLC_VAR_STRING ) )
    {
        msg_Err( p_access, "Valid network information is required." );
        return VLC_ENOMEM;
    }

    config_ChainParse( p_access, SOUT_CFG_PREFIX, ppsz_sout_options, p_access->p_cfg );

    p_sys = vlc_obj_calloc( p_this, 1, sizeof( *p_sys ) );
    if( unlikely( p_sys == NULL ) )
        return VLC_ENOMEM;

    p_sys->gre_src_port = var_InheritInteger(p_access, SOUT_CFG_PREFIX "gre-src-port");
    p_sys->gre_dst_port = var_InheritInteger(p_access, SOUT_CFG_PREFIX "gre-dst-port");
    if (p_sys->gre_dst_port % 2 != 0) {
        msg_Err( p_access, "GRE destination port must be an even number." );
        return VLC_ENOMEM;
    }

    p_sys->i_packet_size = var_InheritInteger(p_access, SOUT_CFG_PREFIX "payload-size" );
    p_sys->p_pktbuffer = block_Alloc( p_sys->i_packet_size );
    if( unlikely(p_sys->p_pktbuffer == NULL) )
        goto failed;

    int i_rist_profile = var_InheritInteger(p_access, SOUT_CFG_PREFIX RIST_PARAM_PROFILE);
    if (rist_client_create(&p_sys->client_ctx, i_rist_profile) != 0) {
        msg_Err( p_access, "Could not create rist client context\n");
        goto failed;
    }

    char *psz_stream_name = NULL;
    psz_stream_name = var_InheritString( p_access, SOUT_CFG_PREFIX RIST_PARAM_CNAME );
    if ( psz_stream_name != NULL && psz_stream_name[0] != '\0')
    {
        if (rist_client_set_cname(p_sys->client_ctx, psz_stream_name, strlen(psz_stream_name)) != 0) {
            msg_Err( p_access, "Could not set the cname\n");
            free(psz_stream_name);
            goto failed;
        }
        free( psz_stream_name );
    }

    uint64_t now;
    struct timeval time;
    gettimeofday(&time, NULL);
    now = time.tv_sec * 1000000;
    now += time.tv_usec;
    uint32_t adv_flow_id = (uint32_t)(now >> 16);
    // It must me an even number
    adv_flow_id &= ~(1UL << 0);

    int i_verbose_level = var_InheritInteger(p_access, SOUT_CFG_PREFIX RIST_PARAM_VERBOSE_LEVEL);
    if(rist_client_init(p_sys->client_ctx, adv_flow_id, i_verbose_level, NULL, NULL, NULL) < 0) {
        msg_Err( p_access, "Could not initialize rist client\n");
        goto failed;
    }

    /* Enable encryption */
    char *psz_shared_secret = NULL;
    psz_shared_secret = var_InheritString( p_access, SOUT_CFG_PREFIX RIST_PARAM_SHARED_SECRET );
    if (psz_shared_secret != NULL && psz_shared_secret[0] != '\0') {
        int i_key_size = var_InheritInteger(p_access, SOUT_CFG_PREFIX RIST_PARAM_ENCRYPTION_TYPE);
        if (i_key_size) {
            if (rist_client_encrypt_enable(p_sys->client_ctx, psz_shared_secret, i_key_size) == -1) {
                msg_Err( p_access, "Could not enable encryption\n");
                free(psz_shared_secret);
                goto failed;
            }
        }
        free(psz_shared_secret);
    }

    int i_recovery_length = var_InheritInteger( p_access, SOUT_CFG_PREFIX RIST_PARAM_BUFFER_SIZE );
    int i_rist_reorder_buffer = var_InheritInteger( p_access, SOUT_CFG_PREFIX RIST_PARAM_REORDER_BUFFER );
    int i_rist_retry_interval = var_InheritInteger( p_access, SOUT_CFG_PREFIX RIST_PARAM_RETRY_INTERVAL );
    int i_rist_max_retries = var_InheritInteger( p_access, SOUT_CFG_PREFIX RIST_PARAM_MAX_RETRIES );
    int i_rist_max_bitrate = var_InheritInteger(p_access, SOUT_CFG_PREFIX RIST_PARAM_MAX_BITRATE);
    int i_multipeer_mode = var_InheritInteger(p_access, SOUT_CFG_PREFIX "multipeer-mode");

    msg_Info( p_access, "Setting retry buffer to %d ms", i_recovery_length );

    struct rist_peer_config peer_config = {
        .address = NULL,
        .gre_dst_port = p_sys->gre_dst_port + 1,
        .recovery_mode = RIST_RECOVERY_MODE_TIME,
        .recovery_maxbitrate = (uint32_t)i_rist_max_bitrate,
        .recovery_maxbitrate_return = 0,
        .recovery_length_min = (uint32_t)i_recovery_length,
        .recovery_length_max = (uint32_t)i_recovery_length,
        .recover_reorder_buffer = (uint32_t)i_rist_reorder_buffer,
        .recovery_rtt_min = (uint32_t)i_rist_retry_interval,
        .recovery_rtt_max = (uint32_t)i_rist_retry_interval,
        .weight = (uint32_t)i_multipeer_mode,
        .bufferbloat_mode = RIST_BUFFER_BLOAT_MODE_NORMAL,
        .bufferbloat_limit = 7,
        .bufferbloat_hard_limit = (uint32_t)i_rist_max_retries
    };

    addr[0] = strdup(p_access->psz_path);
    addr[1] = var_InheritString( p_access, SOUT_CFG_PREFIX RIST_PARAM_URL2 );
    addr[2] = var_InheritString( p_access, SOUT_CFG_PREFIX RIST_PARAM_URL3 );
    addr[3] = var_InheritString( p_access, SOUT_CFG_PREFIX RIST_PARAM_URL4 );
    bool b_peer_alive = false;
    for (size_t i = 0; i < RIST_PEER_COUNT; i++) {
        if (&addr[i] == NULL) {
            continue;
        }
        else if (addr[i][0] == '\0') {
            free(addr[i]);
            continue;
        }
        peer_config.address = addr[i];
        struct rist_peer *peer;
        if (rist_client_add_peer(p_sys->client_ctx, &peer_config, &peer) == -1) {
            msg_Err( p_access, "Could not add peer connector to client for %s", addr[i] );
            free(addr[i]);
            continue;
        }
        else {
            b_peer_alive = true;
        }
        free(addr[i]);
    }
    if (!b_peer_alive)
        goto failed;

    if (rist_client_start(p_sys->client_ctx) == -1) {
        msg_Err( p_access, "Could not start rist client\n");
        goto failed;
    }

    p_access->p_sys = p_sys;
    p_access->pf_write = Write;
    p_access->pf_control = Control;

    return VLC_SUCCESS;

failed:
    rist_client_destroy(p_sys->client_ctx);
    p_sys->client_ctx = NULL;
    return VLC_EGENERIC;
}

#define SRC_PORT_TEXT N_("GRE Source Port")
#define SRC_PORT_LONGTEXT N_( \
    "Source port to be used inside the reduced-mode of the main profile" )

#define DST_PORT_TEXT N_("GRE Destination Port")
#define DST_PORT_LONGTEXT N_( \
    "Destination port to be used inside the reduced-mode of the main profile" )

/* The default target payload size */
#define RIST_DEFAULT_TARGET_PAYLOAD_SIZE 1316

/* Multipeer mode */
#define RIST_DEFAULT_MULTIPEER_MODE 0
#define RIST_MULTIPEER_MODE_TEXT N_("Multipeer mode")
#define RIST_MULTIPEER_MODE_LONGTEXT N_( \
    "This allows you to select between duplicate or load balanced modes when " \
    "sending data to multiple peers (several network paths)" )
static const int multipeer_mode_type[] = { 0, 5, };
static const char *const multipeer_mode_type_names[] = {
    N_("Duplicate"), N_("Load balanced"),
};

/* Module descriptor */
vlc_module_begin()

    set_shortname( N_("RIST") )
    set_description( N_("RIST stream output") )
    set_category( CAT_SOUT )
    set_subcategory( SUBCAT_SOUT_ACO )

    add_integer( SOUT_CFG_PREFIX "payload-size", RIST_DEFAULT_TARGET_PAYLOAD_SIZE,
            N_("RIST target payload size (bytes)"), NULL, true )
    add_integer( SOUT_CFG_PREFIX "gre-src-port", RIST_DEFAULT_SRC_PORT,
            SRC_PORT_TEXT, SRC_PORT_LONGTEXT, true )
    add_integer( SOUT_CFG_PREFIX "gre-dst-port", RIST_DEFAULT_DST_PORT,
            DST_PORT_TEXT, DST_PORT_LONGTEXT, true )
    add_integer( SOUT_CFG_PREFIX "multipeer-mode", RIST_DEFAULT_MULTIPEER_MODE,
            RIST_MULTIPEER_MODE_TEXT, RIST_MULTIPEER_MODE_LONGTEXT, true )
        change_integer_list( multipeer_mode_type, multipeer_mode_type_names )
    add_string( RIST_PARAM_URL2, NULL, RIST_URL2_TEXT, NULL, true )
    add_string( RIST_PARAM_URL3, NULL, RIST_URL3_TEXT, NULL, true )
    add_string( RIST_PARAM_URL4, NULL, RIST_URL4_TEXT, NULL, true )
    add_integer( SOUT_CFG_PREFIX RIST_PARAM_MAX_BITRATE, RIST_DEFAULT_MAX_BITRATE,
            RIST_MAX_BITRATE_TEXT, RIST_MAX_BITRATE_LONGTEXT, true )
    add_integer( SOUT_CFG_PREFIX RIST_PARAM_RETRY_INTERVAL, RIST_DEFAULT_RETRY_INTERVAL, 
        RIST_RETRY_INTERVAL_TEXT, NULL, true )
    add_integer( SOUT_CFG_PREFIX RIST_PARAM_REORDER_BUFFER, RIST_DEFAULT_REORDER_BUFFER, 
        RIST_REORDER_BUFFER_TEXT, NULL, true )
    add_integer( SOUT_CFG_PREFIX RIST_PARAM_MAX_RETRIES, RIST_DEFAULT_MAX_RETRIES, 
        RIST_MAX_RETRIES_TEXT, NULL, true )
    add_integer( SOUT_CFG_PREFIX RIST_PARAM_VERBOSE_LEVEL, RIST_DEFAULT_VERBOSE_LEVEL,
            RIST_VERBOSE_LEVEL_TEXT, RIST_VERBOSE_LEVEL_LONGTEXT, true )
        change_integer_list( verbose_level_type, verbose_level_type_names )
    add_integer( SOUT_CFG_PREFIX RIST_PARAM_BUFFER_SIZE, RIST_DEFAULT_BUFFER_SIZE,
            BUFFER_TEXT, BUFFER_LONGTEXT, true )
    add_string( SOUT_CFG_PREFIX RIST_PARAM_CNAME, NULL, RIST_CNAME_TEXT, 
            RIST_CNAME_LONGTEXT, true )
    add_integer( SOUT_CFG_PREFIX RIST_PARAM_PROFILE, RIST_DEFAULT_PROFILE,
            RIST_PROFILE_TEXT, RIST_PROFILE_LONGTEXT, true )
    add_password( SOUT_CFG_PREFIX RIST_PARAM_SHARED_SECRET, "",
            RIST_SHARED_SECRET_TEXT, NULL )
    add_integer( SOUT_CFG_PREFIX RIST_PARAM_ENCRYPTION_TYPE, RIST_DEFAULT_ENCRYPTION_TPE,
            RIST_ENCRYPTION_TYPE_TEXT, NULL, true )
        change_integer_list( rist_encryption_type, rist_encryption_type_names )

    set_capability( "sout access", 0 )
    add_shortcut( "rist", "tr06" )

    set_callbacks( Open, Close )

vlc_module_end ()
